==> lhcsmapi-1.3.300.tar.gz: lhcsmapi-1.3.300/lhcsmapi/analysis/ItCircuitAnalysis.py <==
import warnings
from typing import Dict
import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression

from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder, create_title
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.reference.Reference import Reference


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '%s\n' % message


warnings.formatwarning = warning_on_one_line


class ItCircuitAnalysis(CircuitAnalysis):
    """Class for analysis of IT (Inner Triplet) quadrupole circuits
    """

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct: Dict[str, int], circuit_name='') -> None:
        """ Method creating timestamp table as well as checking the timing between considered timestamps.

        :param timestamp_dct: a dictionary from timestamp name to its value (it is assumed that does not contain NaN)
        :type timestamp_dct: Dict[str, int]
        :param circuit_name: name of a circuit
        :type circuit_name: str
        """
        super().create_timestamp_table(timestamp_dct)

        if ('QDS_A' not in timestamp_dct) or np.isnan(timestamp_dct['QDS_A']):
            warnings.warn('QDS Board A not present in PM buffer!', stacklevel=2)
        elif ('QDS_B' not in timestamp_dct) or np.isnan(timestamp_dct['QDS_B']):
            warnings.warn('QDS Board B not present in PM buffer!', stacklevel=2)
        else:
            if timestamp_dct['QDS_B'] - timestamp_dct['QDS_A'] != int(1e6):
                warnings.warn('QDS PM events (board A and B) are not synchronized to 1 ms!')

        if ('QDS_A' in timestamp_dct) and ('FGC_RQX' in timestamp_dct):
            if abs(timestamp_dct['QDS_A'] - timestamp_dct['FGC_RQX']) > int(20e6):
                warnings.warn('QDS_A and FGC PM events are not synchronized to +/- 20 ms!')
        else:
            warnings.warn('Either QDS_A or FGC PM event is missing. '
                          'Can not check synchronization of QDS and FGC timestamps!')

    # Quench
    @staticmethod
    def find_time_of_quench(i_ref_df: pd.DataFrame) -> float:
        """ Method finding the start of a quench as one time stamp before the drop of I_REF to 0 A around t = 0 s;
        +/- 2 s.

        :param i_ref_df: reference current of the power converter
        :type i_ref_df: pd.DataFrame
        :return: one time stamp before the drop of I_REF to 0 A
        """
        t_i_ref_min = i_ref_df[(i_ref_df.index > -2) & (i_ref_df.index < 2)].idxmin().values[0]
        return i_ref_df[i_ref_df.index < t_i_ref_min].index[-1]

    @staticmethod
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, plot of I_MEAS_Q1-Q3 skipped!')
    def plot_magnet_current(circuit_name: str, timestamp_fgc: int, i_meas_q1_df: pd.DataFrame,
                            i_meas_q2_df: pd.DataFrame, i_meas_q3_df: pd.DataFrame) -> None:
        """ Method plotting current through inner triplet quadrupole magnets Q1-Q3. The magnet current is calculated
        from the main power converter current.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qds: timestamp of the QDS PM event (for the title)
        :type timestamp_qds: int
        :param i_meas_q1_df: current of the first magnet calculated with respective I_MEAS signals
        :param i_meas_q2_df: current of the second magnet calculated with respective I_MEAS signals
        :param i_meas_q3_df: current of the third magnet calculated with respective I_MEAS signals
        """
        title = create_title(circuit_name, timestamp_fgc, ['I_MEAS_Q1', 'I_MEAS_Q2', 'I_MEAS_Q3'])
        PlotBuilder().with_signal([i_meas_q1_df, i_meas_q2_df, i_meas_q3_df], title=title, figsize=(20, 7), grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_xlim((-1, 2)) \
            .plot()

    @check_dataframe_empty(mode='all', warning='All DataFrames are empty, search for the quench origin is skipped!')
    def find_quench_origin(self, u_res_q1_df: pd.DataFrame, u_res_q2_df: pd.DataFrame,
                           u_res_q3_df: pd.DataFrame) -> None:
        """ Method finding an origin of a quench in Q1-Q3 magnets. First it finds the moment when the detection voltage
        reaches the threshold of 0.1 V. If the threshold is not reached, then the last index of the signal is returned.
        The quench origin is selected as the first magnet to reach the threshold.
        The name of the quench origin is printed out and saved to the results table.

        :param u_res_q1_df: quench detection voltage of the first magnet
        :type u_res_q1_df: pd.DataFrame
        :param u_res_q1_df: quench detection voltage of the second magnet
        :type u_res_q1_df: pd.DataFrame
        :param u_res_q1_df: quench detection voltage of the third magnet
        :type u_res_q1_df: pd.DataFrame
        """
        def find_time_quench_detection(u_res_df: pd.DataFrame) -> float:
            # Take absolute value
            u_res_abs_df = abs(u_res_df)

            # Find first index of argmax (0.1 V)
            if u_res_abs_df[u_res_abs_df >= 0.1].dropna().empty:
                return u_res_abs_df.index[-1]
            else:
                return u_res_abs_df[u_res_abs_df >= 0.1].dropna().index[0]

        t_u_max_q1 = find_time_quench_detection(u_res_q1_df)
        t_u_max_q2 = find_time_quench_detection(u_res_q2_df)
        t_u_max_q3 = find_time_quench_detection(u_res_q3_df)

        # Put into a dict
        q_to_t_u_max = {'Q1': t_u_max_q1, 'Q2': t_u_max_q2, 'Q3': t_u_max_q3}

        # Sort the dictionary based on value
        q_to_t_u_max_sorted = {k: v for k, v in sorted(q_to_t_u_max.items(), key=lambda item: item[1])}

        # Quench origin is the first element in the dictionary sorted in the ascending order
        quench_origin = list(q_to_t_u_max_sorted.keys())[0]
        print('Quench origin is %s.' % quench_origin)
        if self.results_table is not None:
            self.results_table['Quench Origin'] = quench_origin

    @staticmethod
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, plot of U_RES_Q1-Q3 skipped!')
    def plot_u_res_q1_q2_q3(circuit_name: str, timestamp_qds: int, u_res_q1_df: pd.DataFrame, u_res_q2_df: pd.DataFrame,
                            u_res_q3_df: pd.DataFrame) -> None:
        """Method plotting U_RES voltage from QPS for Q1, Q2, Q3. A yellow-green band for the detection threshold is
        displayed.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qds: timestamp of the QDS PM event (for the title)
        :type timestamp_qds: int
        :param u_res_q1_df: U_RES voltage from QPS for Q1 magnet
        :type u_res_q1_df: pd.DataFrame
        :param u_res_q2_df: U_RES voltage from QPS for Q2 magnet
        :type u_res_q2_df: pd.DataFrame
        :param u_res_q3_df: U_RES voltage from QPS for Q3 magnet
        :type u_res_q3_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_qds, ['U_RES_Q1', 'U_RES_Q2', 'U_RES_Q3'])
        PlotBuilder().with_signal([u_res_q1_df, u_res_q2_df, u_res_q3_df], title=title, figsize=(15, 7), grid=True) \
            .with_ylabel(ylabel='U, [V]') \
            .with_xlim((-0.5, 0.5)) \
            .with_axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen') \
            .plot()

    def create_mp3_results_table(self):
        """ Method creating an MP3-compatible quench database row for an IT circuit

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='IT', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=self.results_table.index)

        # Union of MP3 and results table columns
        columns_lhcsm = Reference.read_quench_database_columns(circuit_type='IT', table_type='LHCSM')
        same_columns = list(set(columns_mp3).intersection(set(columns_lhcsm)))

        mp3_table[same_columns] = self.results_table[same_columns]

        # Fill non-matching column names
        max_i_earth_df = self.results_table[['max_I_EARTH_RQX', 'max_I_EARTH_RTQX1', 'max_I_EARTH_RTQX2']]
        idx_abs_min = abs(max_i_earth_df.min()).idxmax()
        idx_max = max_i_earth_df.max().idxmax()
        mp3_table['Earth Current'] = max_i_earth_df[idx_max].max() if max_i_earth_df[idx_max].max() > abs(
            max_i_earth_df[idx_abs_min].min()) else max_i_earth_df[idx_abs_min].min()

        origin = self.results_table['Quench Origin'].values[0]
        if isinstance(origin, str):
            mp3_table['Magnet MIITs'] = self.results_table['MITTs_' + origin]
            mp3_table['dV/dt'] = self.results_table['dU_RES_dt_' + origin]

        mp3_table['DCCT MIITs'] = self.results_table['MITTs_DCCT_RQX']

        mp3_table['Person who analyzed'] = self.results_table['performed_by']

        return mp3_table

    @staticmethod
    def calculate_resistance(i_meas_feature_df, u_mag_feature_df):
        pd.set_option('mode.chained_assignment', None)
        i_meas_feature_df = i_meas_feature_df[['device', 'class', 'mean', 'std']]

        i_rqx_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RQX')]
        i_rtqx1_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RTQX1')]
        i_rtqx2_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('RTQX2')]

        # Calculate I_Q1
        I_Q1_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0] + \
                 i_rtqx1_df[i_rtqx1_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q1', 1, I_Q1_1, 0]

        I_Q1_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0] + \
                 i_rtqx1_df[i_rtqx1_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q1', 2, I_Q1_2, 0]

        # Calculate I_Q2
        I_Q2_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0] + \
                 i_rtqx2_df[i_rtqx2_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q2', 1, I_Q2_1, 0]

        I_Q2_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0] + \
                 i_rtqx2_df[i_rtqx2_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q2', 1, I_Q2_2, 0]

        # Calculate I_Q3
        I_Q3_1 = i_rqx_df[i_rqx_df['class'] == 1]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q3', 1, I_Q3_1, 0]

        I_Q3_2 = i_rqx_df[i_rqx_df['class'] == 2]['mean'].values[0]
        i_meas_feature_df.loc[len(i_meas_feature_df)] = ['I_Q3', 2, I_Q3_2, 0]

        i_q_feature_df = i_meas_feature_df[i_meas_feature_df['device'].str.contains('I_Q')]
        pd.set_option('mode.chained_assignment', 'warn')

        res = {}
        u_mag_feature_df = u_mag_feature_df[(u_mag_feature_df['class'] > 0) & (u_mag_feature_df['class'] < 3)]
        for magnet_name in ['Q1', 'Q2', 'Q3']:
            u_mean = u_mag_feature_df[u_mag_feature_df['nxcals_variable_name'].str.contains(magnet_name)].sort_values(
                by='class')['mean'].values
            i_mean = i_q_feature_df[i_q_feature_df['device'] == 'I_%s' % magnet_name].sort_values(by='class')[
                'mean'].values
            res['R_%s' % magnet_name] = abs(LinearRegression().fit(i_mean.reshape(-1, 1), u_mean).coef_[0])

        return pd.DataFrame({'R_RES': res})

==> lhcsmapi-1.4.58.tar.gz: lhcsmapi-1.4.58/lhcsmapi/analysis/ItCircuitAnalysis.py <==
import warnings
from typing import Dict

import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.analysis.busbar.BusbarResistanceAnalysis import ItBusbarResistanceAnalysis
from lhcsmapi.analysis.dfb.DfbAnalysis import DfbAnalysis
from lhcsmapi.analysis.pc.PcAnalysis import PcAnalysis
from lhcsmapi.analysis.qds.QdsAnalysis import QdsAnalysis
from lhcsmapi.analysis.qh.QuenchHeaterAnalysis import QuenchHeaterAnalysis
from lhcsmapi.analysis.warnings import warning_on_one_line
from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder, create_title
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.reference.Reference import Reference


warnings.formatwarning = warning_on_one_line


class ItCircuitAnalysis(PcAnalysis, QdsAnalysis, QuenchHeaterAnalysis, ItBusbarResistanceAnalysis, DfbAnalysis,
                        CircuitAnalysis):
    """Class for analysis of IT (Inner Triplet) quadrupole circuits

    """

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct: Dict[str, int], circuit_name='') -> None:
        """ Method creating timestamp table as well as checking the timing between considered timestamps.

        :param timestamp_dct: a dictionary from timestamp name to its value (it is assumed that does not contain NaN)
        :type timestamp_dct: Dict[str, int]
        :param circuit_name: name of a circuit
        :type circuit_name: str
        """
        super().create_timestamp_table(timestamp_dct)

        if np.isnan(timestamp_dct.get('QDS_A', np.nan)) and np.isnan(timestamp_dct.get('QDS_B', np.nan)):
            warnings.warn('Neither QDS board A or B are not present in the PM buffer. No QH fired, it is not an FPA!',
                          stacklevel=2)
        elif np.isnan(timestamp_dct.get('QDS_A', np.nan)):
            warnings.warn('QDS Board A not present in PM buffer!', stacklevel=2)
        elif np.isnan(timestamp_dct.get('QDS_B', np.nan)):
            warnings.warn('QDS Board B not present in PM buffer!', stacklevel=2)
        else:
            if timestamp_dct['QDS_B'] - timestamp_dct['QDS_A'] != int(1e6):
                warnings.warn('QDS PM events (board A and B) are not synchronized to 1 ms!')

        if ('QDS_A' in timestamp_dct) and ('FGC_RQX' in timestamp_dct):
            if abs(timestamp_dct['QDS_A'] - timestamp_dct['FGC_RQX']) > int(20e6):
                warnings.warn('QDS_A and FGC PM events are not synchronized to +/- 20 ms!')
        else:
            warnings.warn('Either QDS_A or FGC PM event is missing. '
                          'Can not check synchronization of QDS and FGC timestamps!')

    # Quench
    @staticmethod
    def find_time_of_quench(i_ref_df: pd.DataFrame) -> float:
        """ Method finding the start of a quench as one time stamp before the drop of I_REF to 0 A around t = 0 s;
        +/- 2 s.

        :param i_ref_df: reference current of the power converter
        :type i_ref_df: pd.DataFrame
        :return: one time stamp before the drop of I_REF to 0 A
        """
        t_i_ref_min = i_ref_df[(i_ref_df.index > -2) & (i_ref_df.index < 2)].idxmin().values[0]
        return i_ref_df[i_ref_df.index < t_i_ref_min].index[-1]

    @staticmethod
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, plot of I_MEAS_Q1-Q3 skipped!')
    def plot_magnet_current(circuit_name: str, timestamp_fgc: int, i_meas_q1_df: pd.DataFrame,
                            i_meas_q2_df: pd.DataFrame, i_meas_q3_df: pd.DataFrame) -> None:
        """ Method plotting current through inner triplet quadrupole magnets Q1-Q3. The magnet current is calculated
        from the main power converter current.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qds: timestamp of the QDS PM event (for the title)
        :type timestamp_qds: int
        :param i_meas_q1_df: current of the first magnet calculated with respective I_MEAS signals
        :param i_meas_q2_df: current of the second magnet calculated with respective I_MEAS signals
        :param i_meas_q3_df: current of the third magnet calculated with respective I_MEAS signals
        """
        title = create_title(circuit_name, timestamp_fgc, ['I_MEAS_Q1', 'I_MEAS_Q2', 'I_MEAS_Q3'])
        PlotBuilder().with_signal([i_meas_q1_df, i_meas_q2_df, i_meas_q3_df], title=title, figsize=(13, 6.5), grid=True) \
            .with_ylabel(ylabel='I, [A]') \
            .with_xlim((-1, 2)) \
            .plot()

    @staticmethod
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, plot of U_RES_Q1-Q3 skipped!')
    def plot_u_res_q1_q2_q3(circuit_name: str, timestamp_qds: int, u_res_q1_df: pd.DataFrame, u_res_q2_df: pd.DataFrame,
                            u_res_q3_df: pd.DataFrame) -> None:
        """Method plotting U_RES voltage from QPS for Q1, Q2, Q3. A yellow-green band for the detection threshold is
        displayed.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qds: timestamp of the QDS PM event (for the title)
        :type timestamp_qds: int
        :param u_res_q1_df: U_RES voltage from QPS for Q1 magnet
        :type u_res_q1_df: pd.DataFrame
        :param u_res_q2_df: U_RES voltage from QPS for Q2 magnet
        :type u_res_q2_df: pd.DataFrame
        :param u_res_q3_df: U_RES voltage from QPS for Q3 magnet
        :type u_res_q3_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_qds, ['U_RES_Q1', 'U_RES_Q2', 'U_RES_Q3'])
        PlotBuilder().with_signal([u_res_q1_df, u_res_q2_df, u_res_q3_df], title=title, figsize=(15, 7), grid=True) \
            .with_ylabel(ylabel='U, [V]') \
            .with_xlim((-0.1, 0.2)) \
            .with_axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen') \
            .plot()

    def create_mp3_results_table(self):
        """ Method creating an MP3-compatible quench database row for an IT circuit

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='IT', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=self.results_table.index)

        mp3_table[columns_mp3] = self.results_table[columns_mp3]

        # Fill non-matching column names
        mp3_table['Timestamp_PIC'] = self.results_table.apply(lambda col: Time.to_string_short(col['timestamp_pic']), axis=1)
        mp3_table['Delta_t(FGC_RQX-PIC)'] = self.results_table.apply(lambda col: 1e-6 * (col['timestamp_fgc_rqx'] - col['timestamp_pic']), axis=1)
        mp3_table['Delta_t(QPS-PIC)'] = self.results_table.apply(lambda col: 1e-6 * (col['timestamp_qds_a'] - col['timestamp_pic']), axis=1)
        mp3_table['dU_QPS/dt'] = self.results_table['dU_RES_dt_%s' % self.results_table.loc[0, 'Quench Origin']]

        return mp3_table
